/*================================================================*/
/* COPYRIGHT:                                                     */
/* ----------                                                     */
/* Copyright (C) International Business Machines Corp., 1996.     */
/*                                                                */
/* DISCLAIMER OF WARRANTIES:                                      */
/* -------------------------                                      */
/* The following [enclosed] code is sample code created by IBM    */
/* Corporation.  This sample code is not part of any standard IBM */
/* product and is provided to you solely for the purpose of       */
/* assisting you in the development of your applications.  The    */
/* code is provided "AS IS", without warranty of any kind.  IBM   */
/* shall not be liable for any damages arising out of your use of */
/* the sample code, even if they have been advised of the         */
/* possibility of such damages.                                   */
/*================================================================*/

#define STRICT
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <stdarg.h>
#include <direct.h>
#include <string.h>

//-----------------------------------------------------------------------
// Pull in the VoiceType SMAPI header file(s)
//-----------------------------------------------------------------------
#include <smapi.h>

#include <dmapi.h>


#include "waveit.h"


LONG APIENTRY WndProc ( HWND, UINT, WPARAM, LONG );

BOOL PASCAL   DlgProc ( HWND, UINT, WPARAM, LPARAM );


//-----------------------------------------------------------------------
// Just some local functions (status info, errors, etc)
//-----------------------------------------------------------------------
static void LogError ( char * msg );
static void LogMessage ( char * msg );
static void ChangeColor ( HWND hwnd );
static void DisplayFirmWord ( char * the_word );
static void DisplayInfirmWords ( char * the_words );

static void FormatAndDisplayWord ( char * text,
                                   int join_left, int prev_join_right,
                                   int capitalize,
                                   int uppercase,
                                   int lowercase );

static int  ReadWavList ( char * the_file );
static int  ShowWavList ( );
static int  SelectWavFile ( int which );
static int	QueryAudioSource ( );


//-----------------------------------------------------------------------
// Assorted functions that make SMAPI calls
//-----------------------------------------------------------------------
static int  ConnectToEngine ( HWND hwnd );
static int  DisconnectFromEngine ( HWND hwnd );
static int  TurnMicOn ( );
static int  TurnMicOff ( );


//-----------------------------------------------------------------------
// Callback functions for SMAPI responses
//-----------------------------------------------------------------------
SmHandler ConnectCB       ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler DisconnectCB    ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler MicOnCB         ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler MicOffCB        ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler EnableVocabCB   ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler RecoNextWordCB  ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler RecoWordCB      ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler RecoTextCB      ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler MicOffReqCB     ( SM_MSG rep, caddr_t client, caddr_t caller );
SmHandler UtteranceCB     ( SM_MSG rep, caddr_t client, caddr_t caller );


static char app_name [ ] = "Sample";

static HWND main_win  = NULL;
static HWND reco_text = NULL;
static HWND list_hwnd = NULL;

static char log_buff [ 255 ];
static char msg_text [ 255 ] = "\0";


static DMHANDLE dm_handle;


static char     curr_user [ 40 ] = "\0";
static char     curr_lang [ 40 ] = "\0";
static char     curr_task [ 40 ] = "\0";

static char     wave_list [ 132 ] = "\0";
static char     wave_files [ 32 ] [255 ];
static int      num_waves = 0;
static int      curr_wave = 0;

static int      came_from_engine = FALSE;

//-----------------------------------------------------------------------
// The main program
//-----------------------------------------------------------------------
int WINAPI WinMain ( HINSTANCE inst, HINSTANCE prev_inst,
                     LPSTR cmd_line, int cmd_show )
{
  HWND      hwnd;
  MSG       msg;
  WNDCLASS  wnd_class;

  if ( ! prev_inst )
  {
    wnd_class.style           = CS_HREDRAW | CS_VREDRAW;
    wnd_class.lpfnWndProc     = WndProc;
    wnd_class.cbClsExtra      = 0;
    wnd_class.cbWndExtra      = 0;
    wnd_class.hInstance       = inst;
    wnd_class.hIcon           = LoadIcon ( inst, MAKEINTRESOURCE ( IDI_SAMPLE ) );
    wnd_class.hCursor         = LoadCursor ( NULL, IDC_ARROW );
    wnd_class.hbrBackground   = GetStockObject ( WHITE_BRUSH );
    wnd_class.lpszMenuName    = app_name;
    wnd_class.lpszClassName   = app_name;

    RegisterClass ( &wnd_class );
  }

  if ( cmd_line [ 0 ] )
  {
    strcpy ( wave_list, cmd_line );
  }

  hwnd = CreateWindow ( app_name, "Wave Maker",
                        WS_OVERLAPPEDWINDOW,
                        CW_USEDEFAULT, CW_USEDEFAULT, 650, 300,
                        NULL, NULL, inst, NULL );

  ShowWindow ( hwnd, cmd_show );

  UpdateWindow ( hwnd );

  while ( GetMessage ( & msg, NULL, 0, 0 ) )
  {
    TranslateMessage ( & msg );

    DispatchMessage ( & msg );
  }

  /*--------------------------------------------------------------------*/
  /* Disconnect from the reco engine                                    */
  /*--------------------------------------------------------------------*/
  DisconnectFromEngine ( hwnd );

  return ( msg.wParam );
}


//-----------------------------------------------------------------------
// Window procedure - handle all of the messages (Windows and speech)
//-----------------------------------------------------------------------
long PASCAL WndProc ( HWND hwnd, UINT msg, WPARAM wParam, LONG lParam )
{
  static HANDLE hInstance;
  static PROC   lpfn;
  HMENU         menu;
  HDC           hdc;
  RECT          rect;
  PAINTSTRUCT   ps;

  switch ( msg )
  {
    case WM_CREATE:
         main_win = hwnd;

         hInstance = ( ( LPCREATESTRUCT ) lParam ) ->hInstance;

         lpfn = MakeProcInstance ( DlgProc, hInstance );

         reco_text = CreateWindow ( "edit", NULL,
                                WS_CHILD | WS_VISIBLE | WS_VSCROLL |
                                WS_BORDER | ES_LEFT | ES_MULTILINE |
                                ES_NOHIDESEL | ES_AUTOVSCROLL,
                                10, 10, 310, 220,
                                hwnd, ( HMENU ) IDB_RECOTEXT, hInstance, NULL );

         list_hwnd = CreateWindow ( "listbox", NULL,
                                   WS_CHILD | WS_VISIBLE | LBS_STANDARD |
                                   WS_HSCROLL | WS_VSCROLL,
                                   325, 10, 310, 225,
                                   hwnd, ( HMENU ) 0,
                                   hInstance, ( LPVOID ) NULL );

         if ( wave_list [ 0 ] )
         {
           ReadWavList ( wave_list );

           ShowWavList ( );
         }

         /*-------------------------------------------------------------*/
         /* Establish a connection to the recognition engine            */
         /*-------------------------------------------------------------*/
         ConnectToEngine ( hwnd );

         return ( 0 );

         break;

    case WM_SETFOCUS:
         SetFocus ( reco_text );
         return ( 0 );
         break;

    case WM_PAINT:
         if ( msg_text [ 0 ] )
         {
           hdc = BeginPaint ( hwnd, & ps );

           SetBkMode ( hdc, TRANSPARENT );

           GetClientRect ( hwnd, & rect );

           DrawText ( hdc, msg_text, -1, & rect,
                      DT_SINGLELINE | DT_BOTTOM | DT_LEFT );

           EndPaint ( hwnd, & ps );
         }
         return ( 0 );
         break;

    case WM_COMMAND:
         menu = GetMenu ( hwnd );

         switch ( wParam )
         {
           case IDM_EXIT:
                SendMessage ( hwnd, WM_CLOSE, 0, 0L );
                return ( 0 );
                break;

           case IDM_MIC_ON:
                curr_wave = 0;
                TurnMicOn ( );
                return ( 0 );
                break;

           case IDM_MIC_OFF:
                TurnMicOff ( );
                return ( 0 );
                break;

           case IDM_WAVELIST:
                if ( DialogBox ( hInstance, "WaveList", hwnd, DlgProc ) )
                {
                  InvalidateRect ( hwnd, NULL, TRUE );
                }
                return ( 0 );
                break;

           case IDM_ABCDEF:
                QueryAudioSource ( );
                return ( 0 );
                break;

           /*-------------------------------------------------------------*/
           /* Call whatever callbacks are defined to process a message    */
           /* coming back from the speech engine.                         */
           /*-------------------------------------------------------------*/
           case IDS_SMAPI_ID:

                SmDispatch ( lParam );
                return ( 0 );
                break;
         }
         break;

    case WM_DESTROY:
         PostQuitMessage ( 0 );
         return ( 0 );
         break;
  }

  return ( DefWindowProc ( hwnd, msg, wParam, lParam ) );
}


//-----------------------------------------------------------------------
// Dialog procedure - handle all of the messages for the dialog
//-----------------------------------------------------------------------
BOOL PASCAL DlgProc ( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam )
{

  switch ( msg )
  {
    case WM_INITDIALOG:
         SetDlgItemText ( hwnd, IDD_WAVELIST, wave_list );
         return ( TRUE );
         break;

    case WM_COMMAND:
         switch ( wParam )
         {
           case IDOK:
                GetDlgItemText ( hwnd, IDD_WAVELIST, wave_list, sizeof ( wave_list ) );

                EndDialog ( hwnd, TRUE );

                ReadWavList ( wave_list );

                ShowWavList ( wave_list );

                return ( TRUE );
                break;

           case IDCANCEL:
                EndDialog ( hwnd, FALSE );
                return ( TRUE );
                break;

         }
  }

  return ( FALSE );   // I didn't do anything - the book says return false,
                      // so I'm returning false..
}


/*=======================================================================*/
/*                                                                       */
/*                                                                       */
/* ASSORTED CALLS                                                        */
/*                                                                       */
/*                                                                       */
/*=======================================================================*/

//-----------------------------------------------------------------------
// Change the color of the client area for visual feedback to the user
//-----------------------------------------------------------------------
static void ChangeColor ( HWND hwnd )
{
  static int colors [ ] = { WHITE_BRUSH, LTGRAY_BRUSH, GRAY_BRUSH,
                            DKGRAY_BRUSH, BLACK_BRUSH };
  static int idx = 0;

  SetClassLong ( hwnd, GCL_HBRBACKGROUND,
                 ( LONG ) GetStockObject ( colors [ idx ] ) );

  InvalidateRect ( hwnd, NULL, TRUE );

  idx = idx + 1;

  if ( idx > 4 ) idx = 0;
}


//-----------------------------------------------------------------------
// Display a word in the text area as firm (not highlighted)
//-----------------------------------------------------------------------
static void DisplayFirmWord ( char * the_word )
{
  SendMessage ( reco_text, EM_REPLACESEL, 0, ( LPARAM ) ( LPCSTR ) the_word );
}


//-----------------------------------------------------------------------
// Format a word (if its a macro, etc) and display it in the text area
//-----------------------------------------------------------------------
static void FormatAndDisplayWord ( char * text,
                                   int join_left, int prev_join_right,
                                   int capitalize, int uppercase, int lowercase )
{
  int i;
  int l;

  /*--------------------------------------------------------------------*/
  /* If this word isn't supposed to be joined left, and the prev word   */
  /* didn't say join right, then insert a space in the text control     */
  /*--------------------------------------------------------------------*/
  if ( ( ! prev_join_right ) && ( ! join_left ) )
  {
    DisplayFirmWord ( " " );
  }


  /*--------------------------------------------------------------------*/
  /* If the capitalize flag is set, make the 1st character uppercase    */
  /*--------------------------------------------------------------------*/
  if ( capitalize )
  {
    text [ 0 ] = toupper ( text [ 0 ] );
  }


  /*--------------------------------------------------------------------*/
  /* If uppercase or lowercase were set, go through the whole string    */
  /* performing the necessary conversions                               */
  /*--------------------------------------------------------------------*/
  l = strlen ( text );

  if ( uppercase )
  {
    for ( i = 0 ; i < l ; i++ )
    {
      text [ i ] = toupper ( text [ i ] );
    }
  }


  if ( lowercase )
  {
    for ( i = 0 ; i < l ; i++ )
    {
      text [ i ] = tolower ( text [ i ] );
    }
  }

  /*--------------------------------------------------------------------*/
  /* Finally, display the resulting string in the text control          */
  /*--------------------------------------------------------------------*/
  DisplayFirmWord ( text );
}


//-----------------------------------------------------------------------
// Display some infirm words in the text are (reverse video)        
//-----------------------------------------------------------------------
static void DisplayInfirmWords ( char * the_words )
{
  int  lc;
  int  ll;
  int  pos;
  int  infirm_start;
  int  infirm_len;
  int  infirm_end;

  infirm_len = strlen ( the_words );

  lc = SendMessage ( reco_text, EM_GETLINECOUNT, ( WPARAM ) 0, ( LPARAM ) 0 );

  lc = lc - 1;

  pos = SendMessage ( reco_text, EM_LINEINDEX, ( WPARAM ) lc, ( LPARAM ) 0 );

  if ( pos == -1 )
  {
    LogError ( "DisplayInfirmWords: EM_LINEINDEX returned -1" );
    return;
  }

  ll = SendMessage ( reco_text, EM_LINELENGTH, ( WPARAM ) pos, ( LPARAM ) 0 );

  infirm_start = pos + ll;

  infirm_end = infirm_start + infirm_len;

  SendMessage ( reco_text, EM_REPLACESEL, 0, ( LPARAM ) ( LPCTSTR ) the_words );

  SendMessage ( reco_text, EM_SETSEL, ( WPARAM ) infirm_start,
                ( LPARAM ) infirm_end );
}


//-----------------------------------------------------------------------
// If the file given is an ini file, extract all of the wave file names
// from it, and store their names in 'wave_list', otherwise just store
// the single filename in 'wave_list'
//-----------------------------------------------------------------------
static int ReadWavList ( char * the_file )
{
  char   sbuff [ 16384 ];
  char * cp;
  int    len;
  int    slen;

  num_waves = 0;

  len = strlen ( the_file );

  if ( ( len > 3 ) && 
       ( ( stricmp ( & the_file [ len - 3 ], "wav" ) == 0 ) ||
         ( stricmp ( & the_file [ len - 3 ], "pcm" ) == 0 ) ) )
  {
    strcpy ( wave_files [ num_waves ], the_file );

    num_waves = num_waves + 1;

    return ( 0 );
  }

  slen = GetPrivateProfileSection ( "WavFiles", sbuff, sizeof ( sbuff ),
                                    the_file );

  cp = sbuff;

  while ( cp < ( sbuff + slen - 1 ) )
  {
    len = strlen ( cp );

    if ( strncmp ( cp, "File=", 5 ) == 0 )
    {
      strcpy ( wave_files [ num_waves ], cp + 5 );

      num_waves = num_waves + 1;
    }

    cp = cp + len + 1;

  }

  curr_wave = 0;

  return ( 0 );
}


//-----------------------------------------------------------------------
// Display the names of the wave files that we're going to process
//-----------------------------------------------------------------------
static int ShowWavList ( )
{
  int i;

  SendMessage ( list_hwnd, LB_RESETCONTENT, 0, 0L );

  for ( i = 0 ; i < num_waves ; i++ )
  {
    SendMessage ( list_hwnd, LB_INSERTSTRING, ( WPARAM ) -1,
                  ( LPARAM ) ( LPCSTR ) wave_files [ i ] );
  }

  SendMessage ( list_hwnd, LB_SETCURSEL, ( WPARAM ) -1, ( LPARAM ) 0 );

  return ( 0 );
}


//-----------------------------------------------------------------------
// Display one of the wav files in the list as 'active'             
//-----------------------------------------------------------------------
static int SelectWavFile ( int which )
{
  SendMessage ( list_hwnd, LB_SETCURSEL, ( WPARAM ) which, ( LPARAM ) 0 );

  return ( 0 );
}


//-----------------------------------------------------------------------
// Generate a temporary filename (that will be used when a wav file is
// converted to a pcm file
//-----------------------------------------------------------------------
static char * MakeTempName ( )
{
  static   char tname [ 255 ];
  static   int first = TRUE;
  char     temple [ 255 ];
  char   * cp;

  if ( first )
  {
    cp = getenv ( "TEMP" );

    if ( cp )
    {
      sprintf ( temple, "%s/vtXXXXXX", cp );

      free ( cp );
    }
    else
    {
      strcpy ( temple, "/vtXXXXXX" );
    }

    cp = mktemp ( temple );

    strcpy ( tname, cp );

    first = FALSE;
  }

  return ( tname );
}


//-----------------------------------------------------------------------
// Store a status message in a global variable for display during 
// processing of WM_PAINT messages
//-----------------------------------------------------------------------
static void LogMessage ( char * msg )
{
  strcpy ( msg_text, msg );

  InvalidateRect ( main_win, NULL, TRUE );
}


//-----------------------------------------------------------------------
// Pop up a message box telling the user of something bad that happened
//-----------------------------------------------------------------------
static void LogError ( char * msg )
{
  MessageBox ( NULL, msg, app_name, MB_OK );
}



/*=======================================================================*/
/*                                                                       */
/*                                                                       */
/* CALLS TO SMAPI                                                        */
/*                                                                       */
/*                                                                       */
/*=======================================================================*/

//-----------------------------------------------------------------------
// Check to see if a return code is good (SM_RC_OK) or not, and display
// an error message as appropriate
//-----------------------------------------------------------------------
#define CheckSmRc(thisfn,callfn,rc,rtn)                        \
  if ( rc != SM_RC_OK )                                        \
  {                                                            \
    sprintf ( log_buff, "%s %s = %d", thisfn, callfn, rc );    \
    LogError ( log_buff );                                     \
                                                               \
    return ( rtn );                                            \
  }                                                            \

#define CheckDmRc(thisfn,callfn,rc,rtn)                        \
  if ( ( rc != DM_ERR_OK ) && ( rc != DM_ERR_MACRO_NOT_FOUND ) ) \
  {                                                            \
    sprintf ( log_buff, "%s %s = %d", thisfn, callfn, rc );    \
    LogError ( log_buff );                                     \
                                                               \
    return ( rtn );                                            \
  }                                                            \

#define CheckSmRcRnw(thisfn,callfn,rc,rtn)                              \
  if ( rc != SM_RC_OK )                                                 \
  {                                                                     \
    sprintf ( log_buff, "%s %s = %d", thisfn, callfn, rc );             \
    LogError ( log_buff );                                              \
                                                                        \
    rc = SmRecognizeNextWord ( SmAsynchronous );                        \
                                                                        \
    if ( rc != SM_RC_OK )                                               \
    {                                                                   \
      sprintf ( log_buff, "%s SmRecognizeNextWord = %d", thisfn, rc );  \
      LogError ( log_buff );                                            \
    }                                                                   \
                                                                        \
    return ( rtn );                                                     \
  }                                                                     \


//-----------------------------------------------------------------------
// Establish a recognition session with the speech engine
//-----------------------------------------------------------------------
static int ConnectToEngine ( HWND hwnd )
{
  SmArg  args [ 10 ];
  int    smc;
  int    rc;
  char   * audio_key = "wav;audwav11.dll;wav;garbage";

  LogMessage ( "Connecting..." );

  /*--------------------------------------------------------------------*/
  /* Initialize SMAPI (the API itself).  The SmNwindowHandle and        */
  /* SmNconnectionId are used by the engine and API to pass messages    */
  /* to this application                                                */
  /*--------------------------------------------------------------------*/
  smc = 0;
  SmSetArg ( args [ smc ], SmNapplicationName, app_name );     smc++;
  SmSetArg ( args [ smc ], SmNwindowHandle,    hwnd );         smc++;
  SmSetArg ( args [ smc ], SmNconnectionId,    IDS_SMAPI_ID ); smc++;

  rc = SmOpen ( smc, args );

  CheckSmRc ( "ConnectToEngine", "SmOpen", rc, -1 );

  /*--------------------------------------------------------------------*/
  /* Now set up callbacks to handle responses for all of the calls we   */
  /* are going to make to the engine.  They are called by the           */
  /* SmDispatch() call in WndProc                                       */
  /*--------------------------------------------------------------------*/
  SmAddCallback ( SmNconnectCallback,           ConnectCB,       NULL );
  SmAddCallback ( SmNdisconnectCallback,        DisconnectCB,    NULL );
  SmAddCallback ( SmNmicOnCallback,             MicOnCB,         NULL );
  SmAddCallback ( SmNmicOffCallback,            MicOffCB,        NULL );
  SmAddCallback ( SmNenableVocabCallback,       EnableVocabCB,   NULL );
  SmAddCallback ( SmNrecognizeNextWordCallback, RecoNextWordCB,  NULL );
  SmAddCallback ( SmNrecognizedWordCallback,    RecoWordCB,      NULL );
  SmAddCallback ( SmNrecognizedTextCallback,    RecoTextCB,      NULL );
  SmAddCallback ( SmNmicOffRequestedCallback,   MicOffReqCB,     NULL );
  SmAddCallback ( SmNutteranceCompletedCallback,UtteranceCB,     NULL );


  /*--------------------------------------------------------------------*/
  /* Finally, connect to the speech engine in recognition mode (versus  */
  /* database mode), specifying the default user, enrollment, and task  */
  /*--------------------------------------------------------------------*/
  smc = 0;
  SmSetArg ( args [ smc ], SmNuserId,    SM_USE_CURRENT );  smc++;
  SmSetArg ( args [ smc ], SmNenrollId,  SM_USE_CURRENT );  smc++;
  SmSetArg ( args [ smc ], SmNtask,      SM_USE_CURRENT );  smc++;
  SmSetArg ( args [ smc ], SmNrecognize, TRUE );            smc++;
  SmSetArg ( args [ smc ], SmNaudioHost, audio_key );       smc++;

  rc = SmConnect ( smc, args, SmAsynchronous );

  CheckSmRc ( "ConnectToEngine", "SmConnect", rc, -2 );

  return ( 0 );
}


//-----------------------------------------------------------------------
// Use DMAPI to get all of the defined dictation macros and add them to
// the 'text' vocabulary so they can be dictated
//-----------------------------------------------------------------------
int LoadDictationMacros ( )
{
  int        rc;
  int        i;
  short      number;
  char   * * names;
  SM_VOCWORD   voc_words [ 500 ];
  SM_VOCWORD * voc_ptrs [ 500 ];

  /*--------------------------------------------------------------------*/
  /* Initialize the dictation macro interface using the info that came  */
  /* back in the connection callback                                    */
  /*--------------------------------------------------------------------*/
  rc = DmOpen ( NULL, curr_user, curr_task, curr_lang, & dm_handle );

  CheckDmRc ( "LoadDictationMacros", "DmOpen", rc, -1 );

  /*--------------------------------------------------------------------*/
  /* Now get a list of all of the dictation macro names so they can be  */
  /* added to the text vocabulary                                       */
  /*--------------------------------------------------------------------*/
  rc = DmQueryMacroNames ( dm_handle, & number, & names );

  CheckDmRc ( "LoadDictationMacros", "DmQueryMacroNames", rc, -2 );


  /*--------------------------------------------------------------------*/
  /* And now add all of the dictation macro names to the text vocab     */
  /*--------------------------------------------------------------------*/
  for ( i = 0 ; i < number; i++ )
  {
    voc_ptrs [ i ] = voc_words + i;

    voc_words [ i ].spelling      = names [ i ];
    voc_words [ i ].spelling_size = strlen ( names [ i ] ) + 1;
    voc_words [ i ].flags         = 0;
  }


  rc = SmAddToVocab ( "text", number, voc_ptrs, SmAsynchronous );


  /*--------------------------------------------------------------------*/
  /* And free up the memory DMAPI allocated to store the names          */
  /*--------------------------------------------------------------------*/
  DmFreeNames ( number, names );

  CheckDmRc ( "LoadDictationMacros", "DmFreeNames", rc, -3 );


  return ( 0 );
}


//-----------------------------------------------------------------------
// Turn the microphone on
//-----------------------------------------------------------------------
static int QueryAudioSource ( )
{
  int    rc;
  SM_MSG reply;
  unsigned long item;
  unsigned long len;
  void     * val;
  char     txt [ 255 ];

  rc = SmQueryBinary ( SM_AUDIO_SOURCE, & reply );

  rc = SmGetBinaryItemValue ( reply, & item, & len, & val );

  sprintf ( txt, "Item: %d, Len: %d, Val: '%s'", item, len, ( char * ) val );

  MessageBox ( NULL, txt, app_name, MB_OK );

  return 0;
}


//-----------------------------------------------------------------------
// Turn the microphone on
//-----------------------------------------------------------------------
static int TurnMicOn ( )
{
  int rc;

  LogMessage ( "Turning mic on" );

  if ( wave_list [ 0 ] )
  {
    char * filename;
    int    len;

    filename = wave_files [ curr_wave ];

    len = strlen ( filename );

    rc = SmSetBinary ( SM_AUDIO_SOURCE,
                       ( short ) strlen ( filename), filename,
                       SmAsynchronous );

    CheckSmRc ( "TurnMicOn", "SmSetBinary", rc, -10 );

    SelectWavFile ( curr_wave );

    curr_wave = curr_wave + 1;
  }

  /*--------------------------------------------------------------------*/
  /* Ask the recognizer to turn the microphone on (start listening)     */
  /*--------------------------------------------------------------------*/
  rc = SmMicOn ( SmAsynchronous );

  CheckSmRc ( "TurnMicOn", "SmMicOn", rc, -1 );

  return ( 0 );
}


//-----------------------------------------------------------------------
// Turn the microphone off
//-----------------------------------------------------------------------
static int TurnMicOff ( )
{
  int rc;

  LogMessage ( "Turning mic off" );

  /*--------------------------------------------------------------------*/
  /* Ask the recognizer to turn the microphone off (stop listening)     */
  /*--------------------------------------------------------------------*/
  rc = SmMicOff ( SmAsynchronous );

  CheckSmRc ( "TurnMicOff", "SmMicOff", rc, -1 );

  return ( 0 );
}


//-----------------------------------------------------------------------
// Disconnect from the recognition engine
//-----------------------------------------------------------------------
static int DisconnectFromEngine ( HWND hwnd )
{
  int rc;

  LogMessage ( "Disconnecting" );

  /*--------------------------------------------------------------------*/
  /* Tell the recognizer we're finished with it (it will close if there */
  /* are no more clients connected to it)                               */
  /*--------------------------------------------------------------------*/
  rc = SmDisconnect ( 0, NULL, SmAsynchronous );

  CheckSmRc ( "DisconnectFromEngine", "SmDisconnect", rc, -1 );


  /*--------------------------------------------------------------------*/
  /* Close down/clean up the dictation macro api                        */
  /*--------------------------------------------------------------------*/
  rc = DmClose ( dm_handle );

  return ( 0 );
}


/*=======================================================================*/
/*                                                                       */
/*                                                                       */
/* C A L L B A C K S                                                     */
/*                                                                       */
/*  These functions are called by SmDispatch in WndProc to handle each   */
/*  of the different messages that the speech engine can send back       */
/*                                                                       */
/*=======================================================================*/

//-----------------------------------------------------------------------
// ConnectCB - called when SmConnect completes (or fails)
//-----------------------------------------------------------------------
SmHandler ConnectCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int     rc;
  HMENU   menu;
  char  * cp;

  SmGetRc ( reply, & rc );

  CheckSmRc ( "ConnectCB", "SmGetRc", rc, -1 );

  menu = GetMenu ( main_win );
  EnableMenuItem ( menu, IDM_MIC_ON, MF_ENABLED );

  LogMessage ( "Connected" );

  /*--------------------------------------------------------------------*/
  /* Turn on audio saving (which also saves alternate words)            */
  /*--------------------------------------------------------------------*/
  // rc = SmSet ( SM_SAVE_AUDIO,TRUE, SmAsynchronous );

  CheckSmRc ( "ConnectCB", "SmSet", rc, -2 );

  /*--------------------------------------------------------------------*/
  /* Enable the text (dictation) vocabulary                             */
  /*--------------------------------------------------------------------*/
  rc = SmEnableVocab ( "text", SmAsynchronous );

  CheckSmRc ( "ConnectCB", "SmEnableVocab", rc, -3 );


  /*--------------------------------------------------------------------*/
  /* Get the user information from the reply structure - it will be     */
  /* needed to load the dictation macros                                */
  /*--------------------------------------------------------------------*/
  rc = SmGetUserId ( reply, & cp );

  CheckSmRc ( "ConnectCB", "SmGetUserId", rc, -4 );

  strcpy ( curr_user, cp );

  rc = SmGetTask ( reply, & cp );

  CheckSmRc ( "ConnectCB", "SmGetTask", rc, -5 );

  strcpy ( curr_task, cp );


  strcpy ( curr_lang, "En_US" );


  /*--------------------------------------------------------------------*/
  /* Load all of the dictation macros                                   */
  /*--------------------------------------------------------------------*/
  LoadDictationMacros ( );


  return ( 0 );
}


//-----------------------------------------------------------------------
// DisconnectCB - called when we've disconnected from the engine
//-----------------------------------------------------------------------
SmHandler DisconnectCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  LogMessage ( "Disconnected" );

  return ( SM_RC_OK );
}


//-----------------------------------------------------------------------
// MicOnCB - called when the engine has turned the mic on (or failed to)
//-----------------------------------------------------------------------
SmHandler MicOnCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int   rc;
  HMENU menu;

  SmGetRc ( reply, & rc );

  CheckSmRc ( "MicOnCB", "SmGetRc", rc, -1 );

  /*--------------------------------------------------------------------*/
  /* When the mic comes on, we call SmRecognizeNextWord to tell the     */
  /* engine to 'go' (start recognizing the audio)                       */
  /*--------------------------------------------------------------------*/
  rc = SmRecognizeNextWord ( SmAsynchronous );

  CheckSmRc ( "MicOnCB", "SmRecognizeNextWord", rc, -2 );

  menu = GetMenu ( main_win );
  EnableMenuItem ( menu, IDM_MIC_ON,  MF_GRAYED );
  EnableMenuItem ( menu, IDM_MIC_OFF, MF_ENABLED );

  LogMessage ( "Microphone On" );

  // DisplayInfirmWords ( "Dictating..." );

  return ( SM_RC_OK );
}


//-----------------------------------------------------------------------
// MicOffCB - called when the mic is turned off (or fails to turn off)
//-----------------------------------------------------------------------
SmHandler MicOffCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int   rc;
  HMENU menu;

  SmGetRc ( reply, & rc );

  CheckSmRc ( "MicOffCB", "SmGetRc", rc, -1 );

  menu = GetMenu ( main_win );
  EnableMenuItem ( menu, IDM_MIC_ON,  MF_ENABLED );
  EnableMenuItem ( menu, IDM_MIC_OFF, MF_GRAYED );

  LogMessage ( "Microphone Off" );

  return ( SM_RC_OK );
}


//-----------------------------------------------------------------------
// UtteranceCB - 
//-----------------------------------------------------------------------
SmHandler UtteranceCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int   rc;

  SmGetRc ( reply, & rc );

  CheckSmRc ( "UtteranceCB", "SmGetRc", rc, -1 );

  LogMessage ( "Utterance Complete" );

  if ( ( curr_wave < num_waves ) && ( came_from_engine ) )
  {
    TurnMicOn ( );
  }
  else
  {
    SelectWavFile ( -1 );
  }

  came_from_engine = FALSE;

  return ( SM_RC_OK );
}


//-----------------------------------------------------------------------
// MicOffReqCB - 
//-----------------------------------------------------------------------
SmHandler MicOffReqCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int   rc;

  LogMessage ( "Microphone Off Reuested" );

  SmGetRc ( reply, & rc );

  CheckSmRc ( "MicOffReqCB", "SmGetRc", rc, -1 );

  LogMessage ( "Microphone Off Reuested" );

  came_from_engine= TRUE;

  TurnMicOff ( );

  return ( SM_RC_OK );
}


//-----------------------------------------------------------------------
// EnableVocabCB - called when the engine has enabled a vocabulary
//-----------------------------------------------------------------------
SmHandler EnableVocabCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int    rc;
  char * vocab_name;

  SmGetRc ( reply, & rc );

  CheckSmRc ( "EnableVocabCB", "SmGetRc", rc, -1 );

  rc = SmGetVocabName ( reply, & vocab_name );

  CheckSmRc ( "EnableVocabCB", "SmGetVocabName", rc, -2 );

  sprintf ( log_buff, "Vocab '%s' enabled", vocab_name );

  LogMessage ( log_buff );

  return ( SM_RC_OK );
}


//-----------------------------------------------------------------------
// RecoNextWordCB - called when the engine has begun recognizing (after
//                  we told it to 'go')
//-----------------------------------------------------------------------
SmHandler RecoNextWordCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int rc;

  SmGetRc ( reply, & rc );

  CheckSmRc ( "RecoNextWordCB", "SmGetRc", rc, -1 );

  return ( SM_RC_OK );
}


//-----------------------------------------------------------------------
// RecoWordCB - called whenever the engine has recognized a command word
//              from one of our defined command vocabularies
//-----------------------------------------------------------------------
SmHandler RecoWordCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int       rc;
  SM_WORD * firm;
  int       num_firm;


  SmGetRc ( reply, & rc );

  CheckSmRcRnw ( "RecoWordCB", "SmGetRc", rc, -1 );


  /*--------------------------------------------------------------------*/
  /* Get whatever recognized commands the engine sent out of the reply  */
  /* structure (there should be only one in this case)                  */
  /*--------------------------------------------------------------------*/
  rc = SmGetFirmWords ( reply, & num_firm, & firm );

  CheckSmRcRnw ( "RecoWordCB", "SmGetFirmWords", rc, -2 );
  

  /*------------------------------------------------------------------*/
  /* Tell the engine to 'go' again                                    */
  /*------------------------------------------------------------------*/
  rc = SmRecognizeNextWord ( SmAsynchronous );

  CheckSmRc ( "RecoWordCB", "SmRecognizeNextWord", rc, -3 );
  
  return ( SM_RC_OK );
}


//-----------------------------------------------------------------------
// RecoTextCB - called whenever the engine has recognized something from 
//              the text (dictation) vocabulary.          
//-----------------------------------------------------------------------
SmHandler RecoTextCB ( SM_MSG reply, caddr_t client, caddr_t caller )
{
  int                rc;
  SM_WORD          * firm;
  int                num_firm;
  SM_WORD          * infirm;
  int                num_infirm;
  int                i;
  char               buffer [ 255 ];
  DM_MACRO_STRUCT  * pdm;
  static int         prev_join_right = TRUE;
  static int         capitalize = TRUE;
  static int         uppercase = FALSE;
  static int         lowercase = FALSE;


  SmGetRc ( reply, & rc );

  CheckSmRc ( "RecoTextCB", "SmGetRc", rc, -1 );


  /*------------------------------------------------------------------*/
  /* Get the firm words out of the reply structure                    */
  /*------------------------------------------------------------------*/
  rc = SmGetFirmWords ( reply, & num_firm, & firm );

  CheckSmRc ( "RecoTextCB", "SmGetFirmWords", rc, -2 );


  /*------------------------------------------------------------------*/
  /* And put whatever words there are in the text window (they should */
  /* really be formatted in some way - gluing punctuation, etc)       */
  /*------------------------------------------------------------------*/
  for ( i = 0 ; i < num_firm ; i++ )
  {
    rc = DmGetMacro ( dm_handle, firm [ i ].spelling, DM_EXPAND_NORMAL, & pdm );

    CheckDmRc ( "RecoTextCB", "DmGetMacro", rc, -1 );

    if ( pdm )
    {
      strcpy ( buffer, pdm->text );

      FormatAndDisplayWord ( pdm->text, pdm->joinleft, prev_join_right,
                             capitalize, uppercase, lowercase );

      prev_join_right = pdm->joinright;

      capitalize = ( pdm->casing == DM_CASE_CAPITAL );
      uppercase = ( pdm->casing == DM_CASE_UPPERCASE );
      lowercase = ( pdm->casing == DM_CASE_LOWERCASE );

      DmFreeMacro ( pdm );
    }
    else
    {
      strcpy ( buffer, firm [ i ].spelling );

      FormatAndDisplayWord ( buffer, FALSE, prev_join_right,
                             capitalize, uppercase, lowercase );

      prev_join_right = FALSE;
      capitalize = FALSE;
      uppercase = FALSE;
      lowercase = FALSE;
    }
  }

  /*------------------------------------------------------------------*/
  /* If there were no firm words, then just delete whatever infirm    */
  /* words there are (otherwise they are left behind)                 */
  /*------------------------------------------------------------------*/
  if ( num_firm == 0 )
  {
    SendMessage ( reco_text, EM_REPLACESEL, ( WPARAM ) 0,
                  ( LPARAM ) ( LPCSTR ) "" );
  }


  /*------------------------------------------------------------------*/
  /* Now process the infirm words (display them in reverse video in   */
  /* the text area)                                                   */
  /*------------------------------------------------------------------*/
  rc = SmGetInfirmWords ( reply, & num_infirm, & infirm );

  CheckSmRc ( "RecoTextCB", "SmGetInfirmWords", rc, -3 );

  if ( num_infirm )
  {
    buffer [ 0 ] = ' ';
    buffer [ 1 ] = '\0';

    for ( i = 0 ; i < num_infirm ; i++ )
    {
      strcat ( buffer, infirm [ i ].spelling );
      strcat ( buffer, " " );
    }

    DisplayInfirmWords ( buffer );
  }


  return ( SM_RC_OK );
}
